#!/usr/bin/env python

# Copyright (C) 2014-2015 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
# lagosh.py - interactive shell for Lagopus vswitch.
# from https://github.com/hibitomo/lago-dsl
#

import cmd
import getopt
import inspect
import json
import os
import pydoc
import re
import select
import socket
import sys


class DSLError(Exception):

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class dsl(object):

    def decode_file(self, args):
        if args == []:
            lines = sys.stdin.read().splitlines()
        else:
            lines = open(args[0], 'r').read().splitlines()
        return self.decode(lines)

    def decode(self, lines):
        result = ''
        topcmd = ''
        for line in lines:
            line = re.sub(r'#.*$', r'', line)
            line = re.sub(r'\n', r'', line)
            line = re.sub(r'^.*(enable|disable)\s*$', r'', line)
            words = line.split()
            if words == []:
                continue
            # top level keywords
            if words[0] != topcmd:
                if topcmd != '':
                    result += '}\n'
                topcmd = words[0]
                result += topcmd + ' {\n'
            words.pop(0)
            # identifier
            if (words[0])[0] != '-':
                result += '\t'
                result += re.sub(r'^:([^:])', r'\1', words.pop(0)) + ' {\n'
                tab = '\t'
            else:
                tab = ''
            # params
            fmt = ''
            while words != []:
                word = re.sub(r'^:([^:])', r'\1', words.pop(0))
                if word == 'create':
                    continue
                if word[0] == '-':
                    word = word.replace('-', '', 1)
                    if fmt != '':
                        result += tab + '\t' + fmt + ';\n'
                    fmt = word
                else:
                    fmt = fmt + ' ' + word
            if fmt != '':
                result += tab + '\t' + fmt + ';\n'
            if tab != '':
                result += tab + '}\n'
        if topcmd != '':
            result += '}'
        return result

    def encode_file(self, args):
        if args == []:
            lines = sys.stdin.read().splitlines()
        else:
            lines = open(args[0], 'r').read().splitlines()
        return self.encode(lines)

    def encode(self, lines):
        result = ''
        wordlist = []
        for line in lines:
            line = re.sub(r'#.*$', r'', line)
            line = re.sub(r'\n', r'', line)
            wordlist.append(line.split())

        prewords = []
        childwords = []
        bridge = []
        encodedpre = 0
        linecount = 0
        for words in wordlist:
            linecount += 1
            for word in words:
                # add to childwords if normal word
                if word != '{' and not re.search(r';$', word) and word != '}':
                    childwords.append(word)
                    continue
                # set (add) prewords(repeated words) if block begin.
                if word == '{':
                    prewords.append(childwords)
                    if len(prewords) == 2:
                        if prewords[0][0] == 'bridge':
                            bridge.append(prewords[1][0])
                        prewords.append(['create'])
                    for decwords in prewords[encodedpre:]:
                        result += self.encode_id(decwords)
                    encodedpre = len(decwords)
                    childwords = []
                    continue
                # unset (remove) prewords if block end.
                if word == '}':
                    result += '\n'
                    try:
                        if len(prewords) == 3:
                            prewords.pop()
                        prewords.pop()
                    except Exception:
                        raise DSLError('line ' + str(linecount) +
                                       ': Unbalanced parenthesis')
                    childwords = []
                    encodedpre = 0
                    continue
                # ;
                childwords.append(word)
                result += self.encode_op(childwords)
                childwords = []
        if prewords != []:
            raise DSLError('line ' + str(linecount) +
                           ': End of input, unbalanced parenthesis')
        for word in bridge:
            result += 'bridge ' + word + ' enable\n'
        return result

    def encode_id(self, words):
        result = ''
        for word in words:
            result += word + ' '
        return result

    def encode_op(self, words):
        op = words.pop(0)
        result = '-' + op
        for word in words:
            result += ' ' + word
        return re.sub(r';$', ' ', result)


class ds_client(object):

    port = 12345

    def remove_namespace(self, arg):
        if isinstance(arg, dict):
            for a in arg.iterkeys():
                if isinstance(arg[a], unicode):
                    arg[a] = re.sub(r'^:([^:])', r'\1', arg[a])
                elif isinstance(arg[a], dict):
                    self.remove_namespace(arg[a])
                elif isinstance(arg[a], list):
                    for l in arg[a]:
                        self.remove_namespace(l)

    def open(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect(('127.0.0.1', self.port))

    def close(self):
        del self.sock

    def write(self, arg):
        self.sock.sendall(arg)

    def is_readable(self):
        return select.select([self.sock], [], [], 0) == ([self.sock], [], [])

    def read(self):
        data = ''
        while True:
            res = self.sock.recv(8192)
            data += res
            if not self.is_readable():
                try:
                    jdata = json.loads(data)
                    self.remove_namespace(jdata)
                    break
                except ValueError:
                    continue
        return jdata

    def request(self, arg):
        self.write(arg)
        jdata = self.read()
        if jdata['ret'] != 'OK':
            if 'file' in jdata and 'line' in jdata:
                raise DSLError(jdata['file'] + ':' + str(jdata['line']) +
                               ': ' + jdata['ret'] + ': ' + jdata['data'])
            else:
                raise DSLError(jdata['ret'] + ': ' + jdata['data'])
        return jdata['data']

    def lock(self):
        self.request('lock\n')

    def unlock(self):
        self.request('unlock\n')

    def call(self, arg, response=True):
        try:
            self.open()
        except Exception:
            raise
        self.write(arg)
        if response:
            jdata = self.read()
            self.close()
        else:
            self.close()
            return
        if jdata['ret'] != 'OK':
            if 'file' in jdata and 'line' in jdata:
                raise DSLError(jdata['file'] + ':' + str(jdata['line']) +
                               ': ' + jdata['ret'] + ': ' + jdata['data'])
            else:
                raise DSLError(jdata['ret'] + ': ' + jdata['data'])
        try:
            return jdata['data']
        except Exception:
            return


class Cmd(cmd.Cmd):

    pager = False

    def __init__(self, *args, **kwargs):
        cmd.Cmd.__init__(self, *args, **kwargs)

    def onecmd(self, line):
        try:
            return cmd.Cmd.onecmd(self, line)
        except socket.error:
            print('Socket connection refused.  Lagopus is not running?')

    def completenames(self, text, *ignored):
        dotext = 'do_' + text
        return [a[3:] for a in self.get_names()
                if a.startswith(dotext) and a != 'do_shell']

    def precmd(self, line):
        args = line.split()
        line = ''
        cmd = ''
        for a in args:
            begidx = len(line)
            endidx = begidx + len(a) + 1
            if begidx == 0:
                matches = self.completenames(a, line, begidx, endidx)
                if len(matches) == 1:
                    cmd = matches[0]
                else:
                    cmd = a
                line = cmd
            else:
                if cmd == '':
                    compfunc = self.completedefault
                else:
                    try:
                        compfunc = getattr(self, 'complete_' + cmd)
                    except AttributeError:
                        compfunc = self.completedefault
                matches = compfunc(a, line, begidx, endidx)
                if len(matches) == 1:
                    line += ' ' + matches[0]
                else:
                    line += ' ' + a
        return line

    def emptyline(self):
        return

    def output(self, line):
        if self.pager:
            pydoc.pager(line)
        else:
            print(line)

    def cmdloop(self, showpager=False):
        self.pager = showpager
        cmd.Cmd.cmdloop(self)

    def complete_pager(self, text, line, bigidx, endidx):
        return ['on', 'off']

    def do_pager(self, line):
        if line == 'on':
            self.pager = True
        elif line == 'off':
            self.pager = False
        elif line == '':
            if self.pager:
                print('pager is on.')
            else:
                print('pager is off.')
        else:
            print('Argument error.')

    def do_shell(self, line):
        os.system(line)

    def complete_EOF(self, _text, _line, _begidx, _endidx):
        return []

    def do_EOF(self, _line):
        return True

    def do_exit(self, _line):
        return True

    def do_quit(self, _line):
        return True


class Configure(Cmd):

    def __init__(self, *args, **kwargs):
        self._in_onecmd = False
        self.prompt = 'Configure# '
        Cmd.__init__(self, *args, **kwargs)

    def do_set(self, line):
        """Add or modify candidate configuration line."""
        print('not implemented yet')

    def do_unset(self, line):
        """Remove candidate configuration line."""
        print('not implemented yet')

    def complete_show(self, text, line, bigidx, endidx):
        confdir = os.environ.get('HOME') + '/.lagopus.conf.d'
        params = os.listdir(confdir + '/')
        params.remove('.git')
        return [name for name in params if name.startswith(text)]

    def do_show(self, line):
        """Show the configuration.

        usage:
                show
                show file
        """
        confdir = os.environ.get('HOME') + '/.lagopus.conf.d'
        if not os.path.isdir(confdir):
            os.mkdir(confdir)
        args = line.split()
        try:
            conffile = confdir + '/' + args[0]
        except Exception:
            lines = ds_client().call('save\n').splitlines()
            self.output(dsl().decode(lines))
            return
        try:
            self.output(open(conffile).read())
        except Exception:
            print("failed to read " + conffile + "\n")

    def complete_commit(self, text, line, begidx, endidx):
        confdir = os.environ.get('HOME') + '/.lagopus.conf.d'
        params = os.listdir(confdir + '/')
        params.remove('.git')
        return [name for name in params if name.startswith(text)]

    def do_commit(self, line):
        """Commit configuration."""
        confdir = os.environ.get('HOME') + '/.lagopus.conf.d'
        args = line.split()
        try:
            if '/' in args[0]:
                conffile = args[0]
            else:
                conffile = confdir + '/' + args[0]
        except Exception:
            conffile = confdir + '/' + 'lagopus.conf'

        if not os.path.isfile(conffile):
            print('Do not exist configuration')
            return
        # Load candidate config and convert to DSL
        dslfile = conffile + '.dsl'
        f = open(dslfile, 'w')
        try:
            f.write(dsl().encode_file([conffile]))
        except DSLError as e:
            print(e.value)
            f.close()
            return
        f.close()
        # Load DSL.
        ds_client().call('destroy-all-obj\n')
        try:
            ds_client().call('load ' + dslfile + '\n')
        except DSLError as e:
            print(e.value)

    def do_load(self, line):
        """Load DSL

        Load configuration file and commit immediately.
        So far, use raw DSL output.
        """
        conffile = '@PREFIX@/etc/lagopus/lagopus.dsl'
        # Load candidate config and convert to DSL
        # Load DSL.
        ds_client().call('destroy-all-obj\n')
        try:
            ds_client().call('load ' + conffile + '\n')
        except DSLError as e:
            print(e.value)

    def do_save(self, line):
        """Save DSL

        Load configuration file and commit immediately.
        So far, use raw DSL output.
        """
        conffile = '@PREFIX@/etc/lagopus/lagopus.dsl'
        # Load candidate config and convert to DSL
        # Load DSL.
        ds_client().call('save ' + conffile + '\n')

    def complete_edit(self, text, line, bigidx, endidx):
        if inspect.currentframe().f_back.f_code.co_name != 'precmd':
            confdir = os.environ.get('HOME') + '/.lagopus.conf.d'
            params = os.listdir(confdir + '/')
            params.remove('.git')
            return [name for name in params if name.startswith(text)]
        else:
            return []

    def do_edit(self, line):
        """Edit candidate configuration.  Launch text editor."""
        confdir = os.environ.get('HOME') + '/.lagopus.conf.d'
        if not os.path.isdir(confdir):
            os.mkdir(confdir)
        if not os.path.isdir(confdir + '/.git'):
            os.system('cd ' + confdir + '; git init')
        args = line.split()
        try:
            conffile = confdir + '/' + args[0]
        except Exception:
            conffile = confdir + '/' + 'lagopus.conf'
            if not os.path.isfile(conffile):
                # save running DSL if candidate file is not exist.
                lines = ds_client().call('save\n').splitlines()
                f = open(conffile, 'w')
                f.write(dsl().decode(lines))
                f.close()
                # first git commit
                os.system('cd ' + confdir + '; git add ' + conffile)
                os.system('cd ' + confdir +
                          '; git commit --allow-empty-message -m ""')
        # Edit
        editor = os.environ.get('EDITOR', 'vi')
        try:
            st = os.system('cd ' + confdir + ';' + editor + ' ' + conffile)
            if st != 0:
                print('Aborted.')
                return
        except Exception:
            print('Execution failed.')
            return
        # git commit automatically
        os.system('cd ' + confdir + '; git add ' + conffile)
        os.system('cd ' + confdir + '; git commit --allow-empty-message -m ""')

    def do_ls(self, line):
        """Show configuration file contents."""
        confdir = os.environ.get('HOME') + '/.lagopus.conf.d'
        if not os.path.isdir(confdir):
            os.mkdir(confdir)
        os.system('cd ' + confdir + '; /bin/ls ' + line)

    def complete_diff(self, text, line, bigidx, endidx):
        confdir = os.environ.get('HOME') + '/.lagopus.conf.d'
        params = os.listdir(confdir + '/')
        params.remove('.git')
        return [name for name in params if name.startswith(text)]

    def do_diff(self, line):
        """Show differences from previous configuration."""
        confdir = os.environ.get('HOME') + '/.lagopus.conf.d'
        if not os.path.isdir(confdir):
            return
        if not os.path.isdir(confdir + '/.git'):
            return
        args = line.split()
        try:
            conffile = args[-1]
            if len(args) > 0 and os.path.isfile(confdir + '/' + conffile):
                args.pop(-1)
            else:
                raise
        except Exception:
            conffile = 'lagopus.conf'
            if not os.path.isfile(confdir + '/' + conffile):
                return
        if len(args) < 2:
            os.system('cd ' + confdir + '; git diff HEAD^ ' + conffile)
        else:
            os.system('cd ' + confdir + '; git diff ' +
                      ' '.join(map(str, args)) + ' ' + conffile)

    def do_history(self, line):
        """Show differences from previous configuration."""
        confdir = os.environ.get('HOME') + '/.lagopus.conf.d'
        if not os.path.isdir(confdir):
            return
        if not os.path.isdir(confdir + '/.git'):
            return
        args = line.split()
        try:
            conffile = args[0]
        except Exception:
            conffile = 'lagopus.conf'
            if not os.path.isfile(confdir + '/' + conffile):
                return
        os.system('cd ' + confdir + '; git log ' + conffile)


class Topcmd(Cmd):

    def __init__(self, *args, **kwargs):
        self._in_onecmd = False
        self.prompt = 'Lagosh> '
        Cmd.__init__(self, *args, **kwargs)

    def complete_configure(self, text, line, begidx, endidx):
        dotext = 'do_' + text
        return [a[3:] for a in Configure().get_names()
                if a.startswith(dotext) and a != 'do_shell']

    def do_configure(self, line):
        """Manipulate software configuration information."""
        if line == '':
            Configure().cmdloop(showpager=self.pager)
        else:
            Configure().onecmd(line)

    def complete_show(self, text, line, begidx, endidx):
        params = ['bridge', 'channel', 'controller', 'flow', 'group',
                  'mactable', 'interface', 'meter', 'port', 'route', 'version']
        return [name for name in params if name.startswith(text)]

    def subcmd_id_merge(self, subcmd, subcmd_id, res, resc):
        gstats = res[subcmd + 's']
        gconfig = resc[subcmd + 's']
        for gs in gstats:
            for gc in gconfig:
                if gs[subcmd + '-id'] == gc[subcmd + '-id']:
                    stats = gs[subcmd_id + '-stats']
                    config = gc[subcmd_id + 's']
                    for b in stats:
                        for c in config:
                            if c[subcmd_id + '-id'] == b[subcmd_id + '-id']:
                                b.update(c)

    def subcmd_show(self, line, key='bridge', stats='stats', subcmd_id=None):

        if len(line) == 0:
            print('Argument error')
            return

        args = line.split()
        subcmd = args[0]

        if len(args) >= 2:
            req = subcmd + ' ' + args[1] + ' ' + stats + '\n'
            try:
                res = ds_client().call(req)[0]
            except Exception:
                print('invalid keyword: ' + args[1])
                return
            if subcmd_id is not None:
                resc = ds_client().call(subcmd + ' ' + args[1] + '\n')[0]
                self.subcmd_id_merge(subcmd, subcmd_id, res, resc)
            if len(args) == 3:
                try:
                    res = res[args[2]]
                except Exception:
                    print('invalid keyword: ' + args[2])
                    return
            self.output(json.dumps(res, indent=4) + '\n')
        else:
            try:
                odata = []
                data = ds_client().call(key + '\n')
                for ifdata in data:
                    name = ifdata['name']
                    req = subcmd + ' ' + name.encode() + ' ' + stats + '\n'
                    res = ds_client().call(req)[0]
                    if subcmd_id is not None:
                        resc = ds_client().call((subcmd + ' '
                                                 + name.encode() + '\n'))[0]
                        self.subcmd_id_merge(subcmd, subcmd_id, res, resc)
                    try:
                        res['name'] = name
                        res['is-enabled'] = ifdata['is-enabled']
                    except Exception:
                        pass
                    odata.append(res)
                self.output(json.dumps(odata, indent=4) + '\n')
            except TypeError:
                return

    def do_show(self, line):
        """Show statistics.

        Usage
                show bridge
                show flow
                show mactable
                show interface
                show table
        """
        if len(line) == 0:
            print('Argument error')
            return

        args = line.split()
        subcmd = args[0]

        if subcmd == 'bridge':
            self.subcmd_show(line)

        elif subcmd == 'group':
            self.subcmd_show(line, 'bridge', 'stats', 'bucket')

        elif subcmd == 'meter':
            self.subcmd_show(line, 'bridge', 'stats', 'band')

        elif subcmd == 'flow':
            self.subcmd_show(line, 'bridge', '-with-stats')

        elif subcmd in ['interface', 'port']:
            self.subcmd_show(line, subcmd)

        elif subcmd == 'channel':
            data = ds_client().call('channel\n')
            self.output(json.dumps(data, indent=4))

        elif subcmd == 'controller':
            data = ds_client().call('controller\n')
            self.output(json.dumps(data, indent=4))

        elif subcmd == 'mactable':
            odata = []
            data = ds_client().call('bridge\n')
            try:
                for ifdata in data:
                    req = 'mactable ' + ifdata['name'].encode() + '\n'
                    res = ds_client().call(req)[0]
                    odata.append(res)
                self.output(json.dumps(odata, indent=4) + '\n')
            except Exception:
                return

        elif subcmd == 'route':
            data = ds_client().call('route\n')
            self.output(json.dumps(data, indent=4))

        elif subcmd == 'version':
            data = ds_client().call('version\n')
            self.output(json.dumps(data, indent=4))

        else:
            print('Argument error')

    def do_stop(self, line):
        """Stop lagopus process."""
        ds_client().call('shutdown\n', False)

    def do_log(self, line):
        """Show log settings."""
        data = ds_client().call('log\n')
        self.output(json.dumps(data, indent=4))

    def do_telnet(self, line):
        """Telnet to another host."""
        if len(line) == 0:
            print('Argument error')
            return

        args = line.split()
        host = args[0]
        os.system('telnet ' + host)

    def do_ssh(self, line):
        """Connect to another host with secure shell."""
        if len(line) == 0:
            print('Argument error')
            return

        args = line.split()
        host = args[0]
        os.system('ssh ' + host)

    def do_ping(self, line):
        """Ping a remote target."""
        if len(line) == 0:
            print('Argument error')
            return

        args = line.split()
        host = args[0]
        os.system('ping ' + host)

    def do_traceroute(self, line):
        """Trace the route to a remote host."""

        if len(line) == 0:
            print('Argument error')
            return
        args = line.split()
        host = args[0]
        os.system('traceroute ' + host)


def usage():
    print("Usage: " + sys.argv[0] + "[--dsl-decode [file]] "
          "[--dsl-encode [file]]")


def main():
    Topcmd().cmdloop()


if __name__ == "__main__":
    try:
        opts, args = getopt.getopt(sys.argv[1:],
                                   'p:c',
                                   ['dsl-decode=', 'dsl-encode='])
    except getopt.GetoptError:
        usage()
        sys.exit(2)
    for opt, arg in opts:
        if opt == '--dsl-decode':
            print(dsl().decode_file([arg]))
            sys.exit(0)
        elif opt == '--dsl-encode':
            try:
                print(dsl().encode_file([arg]))
            except DSLError as e:
                print(e.value)
            sys.exit(0)
        elif opt == '-p':
            ds_client.port = int(arg)
        elif opt == '-c':
            str = ''
            for word in args:
                str += word + ' '
            Topcmd().onecmd(str)
            sys.exit(0)
    main()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import socket

from neutron_lib import constants as n_const
from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from networking_lagopus.agent import lagosh

LOG = logging.getLogger(__name__)
SOCKET_ISSUE = "Socket connection refused.  Lagopus is not running?\n"


class LagopusCommand(object):

    def _lagosh(self, cmd=None):
        if not cmd:
            return
        lagosh_client = lagosh.ds_client()
        try:
            return lagosh_client.call(cmd)
        except socket.error:
            LOG.debug("_lagosh socket error")
        except lagosh.DSLError as e:
            LOG.debug("_lagosh DSLError cmd: %s, error: %s", cmd, e)

    def show_interfaces(self):
        cmd = "interface\n"
        return self._lagosh(cmd)

    def show_ports(self):
        cmd = "port\n"
        return self._lagosh(cmd)

    def show_bridges(self):
        cmd = "bridge\n"
        return self._lagosh(cmd)

    def show_channels(self):
        cmd = "channel\n"
        return self._lagosh(cmd)

    def show_controllers(self):
        cmd = "controller\n"
        return self._lagosh(cmd)

    @log_helpers.log_method_call
    def create_channel(self, name):
        cmd = "channel %s create -dst-addr 127.0.0.1 -protocol tcp\n" % name
        self._lagosh(cmd)

    @log_helpers.log_method_call
    def create_controller(self, name, channel):
        cmd = ("controller %s create -channel %s -role equal "
               "-connection-type main\n") % (name, channel)
        self._lagosh(cmd)

    @log_helpers.log_method_call
    def create_bridge(self, name, controller, dpid):
        cmd = ("bridge %s create -controller %s -dpid %d "
               "-l2-bridge True -mactable-ageing-time 300 "
               "-mactable-max-entries 8192\n") % (name, controller, dpid)
        self._lagosh(cmd)
        cmd = "bridge %s enable\n" % name
        self._lagosh(cmd)

    # TODO(hichihara): unify create_*_interface
    @log_helpers.log_method_call
    def create_vhost_interface(self, name, device):
        cmd = ("interface %s create -type ethernet-dpdk-phy "
               "-device %s\n") % (name, device)
        self._lagosh(cmd)

    def create_pipe_interface(self, name, device):
        self.create_vhost_interface(name, device)

    @log_helpers.log_method_call
    def create_rawsock_interface(self, name, device):
        cmd = ("interface %s create -type ethernet-rawsock "
               "-device %s\n") % (name, device)
        self._lagosh(cmd)

    @log_helpers.log_method_call
    def create_port(self, port, interface):
        cmd = "port %s create -interface %s\n" % (port, interface)
        self._lagosh(cmd)

    @log_helpers.log_method_call
    def destroy_port(self, port):
        cmd = "port %s destroy\n" % port
        self._lagosh(cmd)

    @log_helpers.log_method_call
    def destroy_interface(self, interface):
        cmd = "interface %s destroy\n" % interface
        self._lagosh(cmd)

    @log_helpers.log_method_call
    def bridge_add_port(self, bridge_name, port_name, ofport):
        cmd = ("bridge %s config -port %s %s\n" %
               (bridge_name, port_name, ofport))
        self._lagosh(cmd)

    @log_helpers.log_method_call
    def bridge_del_port(self, bridge_name, port_name):
        cmd = "bridge %s config -port -%s\n" % (bridge_name, port_name)
        self._lagosh(cmd)

    def find_bridge_port(self, port_id, bridge_name=None):
        if port_id.startswith(n_const.TAP_DEVICE_PREFIX):
            port_id = port_id[len(n_const.TAP_DEVICE_PREFIX):]
        bridges = self.show_bridges()
        for bridge in bridges:
            if bridge_name and bridge["name"] != bridge_name:
                continue
            ports = bridge["ports"]
            for port in ports:
                port_name = port[1:]
                if port_name.startswith(port_id):
                    return bridge["name"], port_name
        return None, None



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants as n_const
from neutron_lib import context
from oslo_log import log as logging

from neutron.agent.linux import interface as n_interface
from neutron.agent.linux import ip_lib
from neutron.agent import rpc as agent_rpc
from neutron.common import topics

from networking_lagopus.agent import rpc as lagopus_rpc

LOG = logging.getLogger(__name__)


class LagopusInterfaceDriver(n_interface.LinuxInterfaceDriver):

    DEV_NAME_PREFIX = 'ns-'

    def __init__(self, conf):
        super(LagopusInterfaceDriver, self).__init__(conf)
        self.context = context.get_admin_context_without_session()
        self.host = self.conf.host
        self.agent_id = 'lagopus-agent-%s' % self.host
        self.plugin_api = agent_rpc.PluginApi(topic=topics.PLUGIN)
        self.lagopus_api = lagopus_rpc.LagopusAgentApi()

    def _get_network_segment(self, port_id):
        details = self.plugin_api.get_device_details(self.context,
                                                     port_id,
                                                     self.agent_id,
                                                     self.host)
        if details.get('physical_network'):
            return {'physical_network': details['physical_network'],
                    'network_type': details['network_type'],
                    'segmentation_id': details['segmentation_id']}

    def _disable_tcp_offload(self, namespace, device_name):
        ip_wrapper = ip_lib.IPWrapper(namespace)
        cmd = ['ethtool', '-K', device_name, 'tx', 'off', 'tso', 'off']
        ip_wrapper.netns.execute(cmd)

    def plug_new(self, network_id, port_id, device_name, mac_address,
                 bridge=None, namespace=None, prefix=None, mtu=None):
        """Plugin the interface."""
        ip = ip_lib.IPWrapper()
        tap_name = device_name.replace(prefix or self.DEV_NAME_PREFIX,
                                       n_const.TAP_DEVICE_PREFIX)
        root_veth, ns_veth = ip.add_veth(tap_name, device_name,
                                         namespace2=namespace)
        root_veth.disable_ipv6()
        ns_veth.link.set_address(mac_address)

        if mtu:
            root_veth.link.set_mtu(mtu)
            ns_veth.link.set_mtu(mtu)
        else:
            LOG.warning("No MTU configured for port %s", port_id)

        root_veth.link.set_up()
        ns_veth.link.set_up()
        self._disable_tcp_offload(namespace, device_name)

        # do first and error check
        segment = self._get_network_segment(port_id)

        self.lagopus_api.plug_rawsock(self.context, tap_name, segment)
        self.plugin_api.update_device_up(self.context, port_id,
                                         self.agent_id, self.host)

    def unplug(self, device_name, bridge=None, namespace=None, prefix=None):
        """Unplug the interface."""
        device = ip_lib.IPDevice(device_name, namespace=namespace)
        tap_name = device_name.replace(prefix or self.DEV_NAME_PREFIX,
                                       n_const.TAP_DEVICE_PREFIX)
        try:
            self.lagopus_api.unplug_rawsock(self.context, tap_name)
            device.link.delete()
            LOG.debug("Unplugged interface '%s'", device_name)
        except RuntimeError:
            LOG.error("Failed unplugging interface '%s'",
                      device_name)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from networking_lagopus._i18n import _

DEFAULT_BRIDGE_MAPPINGS = []

lagopus_opts = [
    cfg.BoolOpt('vhost_mode', default=True,
                help=_("Boot virtual machines with vhost_mode ")),
    cfg.ListOpt('bridge_mappings',
                default=DEFAULT_BRIDGE_MAPPINGS,
                help=_("Comma-separated list of "
                       "<physical_network>:<brige> tuples "
                       "mapping physical network names to the agent's "
                       "node-specific physical network interfaces to be used "
                       "for flat and VLAN networks. All physical networks "
                       "listed in network_vlan_ranges on the server should "
                       "have mappings to appropriate interfaces on each "
                       "agent.")),
    cfg.IPOpt('of_listen_address', default='127.0.0.1',
              help=_("Address to listen on for OpenFlow connections. "
                     "Used only for 'native' driver.")),
    cfg.PortOpt('of_listen_port', default=6633,
                help=_("Port to listen on for OpenFlow connections. "
                       "Used only for 'native' driver.")),
]


cfg.CONF.register_opts(lagopus_opts, "lagopus")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import sys

from oslo_config import cfg
from ryu.base import app_manager
from ryu import cfg as ryu_cfg

from neutron.common import config as common_config
from neutron.common import profiler

from networking_lagopus.common import config  # noqa


LAGOPUS_AGENT_BINARY = 'neutron-lagopus-agent'


def init_ryu_config():
    ryu_cfg.CONF(project='ryu', args=[])
    ryu_cfg.CONF.ofp_listen_host = cfg.CONF.lagopus.of_listen_address
    ryu_cfg.CONF.ofp_tcp_listen_port = cfg.CONF.lagopus.of_listen_port

    # enable Debug log for ryu modules
    # why it is necessary see neutron/common/config.py
    for mod in ('ryu.base.app_manager', 'ryu.controller.controller'):
        logger = logging.getLogger(mod)
        logger.setLevel(logging.DEBUG)


def main():
    common_config.init(sys.argv[1:])
    common_config.setup_logging()
    init_ryu_config()
    profiler.setup(LAGOPUS_AGENT_BINARY, cfg.CONF.host)
    app_manager.AppManager.run_apps([
        'networking_lagopus.ml2.agent.lagopus_ryuapp'
    ])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import eventlet
import os
import sys

from neutron_lib import context
from neutron_lib.utils import helpers
from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_service import loopingcall
from oslo_service import service
from osprofiler import profiler
from ryu.app.ofctl import api as ofctl_api

from neutron.agent import rpc as agent_rpc
from neutron.api.rpc.callbacks import resources
from neutron.common import config as common_config
from neutron.common import rpc as n_rpc
from neutron.common import topics
from neutron.plugins.common import constants as p_constants
from neutron.plugins.ml2.drivers.agent import config as agent_config  # noqa

from networking_lagopus.agent import lagopus_lib
from networking_lagopus.common import config  # noqa


LOG = logging.getLogger(__name__)

LAGOPUS_AGENT_BINARY = 'neutron-lagopus-agent'
AGENT_TYPE_LAGOPUS = 'Lagopus agent'
MAX_WAIT_LAGOPUS_RETRY = 5
OFPP_MAX = 0xffffff00


class LagopusBridge(object):

    def __init__(self, ryu_app, name, dpid, port_data):
        LOG.debug("LagopusBridge: %s %s", name, dpid)
        self.ryu_app = ryu_app
        self.name = name
        self.lagopus_client = lagopus_lib.LagopusCommand()

        self.used_ofport = []
        self.max_ofport = 0
        self.port_mappings = {}

        if port_data:
            for port, ofport in port_data.items():
                # remove ':'
                port_name = port[1:]
                self.port_mappings[port_name] = ofport
                self.used_ofport.append(ofport)
            if self.used_ofport:
                self.max_ofport = max(self.used_ofport)

        LOG.debug("used_ofport: %s, max_ofport: %d",
                  self.used_ofport, self.max_ofport)

        self.dpid = dpid
        self.datapath = self._get_datapath()
        self.install_normal()
        # just for debug
        self.dump_flows()
        return

    def _get_datapath(self):
        # TODO(hichihara): set timeout
        # NOTE: basically it is OK because lagopus is running
        # and dpid exists at this point. so the call shoud be
        # success.
        while True:
            dp = ofctl_api.get_datapath(self.ryu_app, self.dpid)
            if dp is not None:
                return dp
            # lagopus switch dose not establish connection yet.
            # wait a while
            eventlet.sleep(1)

    def install_normal(self):
        ofp = self.datapath.ofproto
        ofpp = self.datapath.ofproto_parser

        actions = [ofpp.OFPActionOutput(ofp.OFPP_NORMAL, 0)]
        instructions = [ofpp.OFPInstructionActions(
                        ofp.OFPIT_APPLY_ACTIONS, actions)]
        msg = ofpp.OFPFlowMod(self.datapath,
                              table_id=0,
                              priority=0,
                              instructions=instructions)
        # TODO(hichihara): error handling
        ofctl_api.send_msg(self.ryu_app, msg)

    def install_vlan(self, vlan_id, port_name):
        ofport = self.port_mappings[port_name]
        ofp = self.datapath.ofproto
        ofpp = self.datapath.ofproto_parser

        # pipe port -> phys port: push vlan, output:1
        match = ofpp.OFPMatch(in_port=ofport)
        vlan_vid = vlan_id | ofp.OFPVID_PRESENT
        actions = [ofpp.OFPActionPushVlan(),
                   ofpp.OFPActionSetField(vlan_vid=vlan_vid),
                   ofpp.OFPActionOutput(1, 0)]
        instructions = [ofpp.OFPInstructionActions(
                        ofp.OFPIT_APPLY_ACTIONS, actions)]
        msg = ofpp.OFPFlowMod(self.datapath,
                              table_id=0,
                              priority=2,
                              match=match,
                              instructions=instructions)
        # TODO(hichihara): error handling
        ofctl_api.send_msg(self.ryu_app, msg)

        # phys port -> pipe port: pop vlan, output:<ofport>
        vlan_vid = vlan_id | ofp.OFPVID_PRESENT
        match = ofpp.OFPMatch(in_port=1, vlan_vid=vlan_vid)
        actions = [ofpp.OFPActionPopVlan(),
                   ofpp.OFPActionOutput(ofport, 0)]
        instructions = [ofpp.OFPInstructionActions(
                        ofp.OFPIT_APPLY_ACTIONS, actions)]
        msg = ofpp.OFPFlowMod(self.datapath,
                              table_id=0,
                              priority=2,
                              match=match,
                              instructions=instructions)
        # TODO(hichihara): error handling
        ofctl_api.send_msg(self.ryu_app, msg)

    def dump_flows(self):
        ofpp = self.datapath.ofproto_parser
        msg = ofpp.OFPFlowStatsRequest(self.datapath)
        reply_cls = ofpp.OFPFlowStatsReply
        # TODO(hichihara): error handling
        result = ofctl_api.send_msg(self.ryu_app, msg, reply_cls=reply_cls,
                                    reply_multi=True)
        LOG.debug("%s flows: %s", self.name, result)

    def get_ofport(self):
        if self.max_ofport < OFPP_MAX:
            self.max_ofport += 1
            self.used_ofport.append(self.max_ofport)
            return self.max_ofport
        for num in range(1, OFPP_MAX + 1):
            if num not in self.used_ofport:
                self.used_ofport.append(num)
                return num

    def free_ofport(self, ofport):
        if ofport in self.used_ofport:
            self.used_ofport.remove(ofport)

    def add_port(self, port_name):
        b, p = self.lagopus_client.find_bridge_port(port_name, self.name)
        if b is not None:
            LOG.debug("port %s is already pluged.", port_name)
            return

        ofport = self.get_ofport()
        self.port_mappings[port_name] = ofport
        self.lagopus_client.bridge_add_port(self.name, port_name, ofport)


@profiler.trace_cls("rpc")
class LagopusManager(object):

    def __init__(self, ryu_app, bridge_mappings):
        self.lagopus_client = lagopus_lib.LagopusCommand()
        self.bridge_mappings = bridge_mappings
        self.ryu_app = ryu_app

        raw_bridges = self._get_init_bridges()

        self.bridges = {}
        name_to_dpid = {}
        bridge_names = bridge_mappings.values()
        for raw_bridge in raw_bridges:
            name = raw_bridge["name"]
            dpid = raw_bridge["dpid"]
            ports = raw_bridge["ports"]
            self.bridges[dpid] = LagopusBridge(ryu_app, name, dpid, ports)
            if name in bridge_names:
                name_to_dpid[name] = dpid

        self.phys_to_dpid = {}
        for phys_net, name in bridge_mappings.items():
            if name not in name_to_dpid:
                LOG.error("Bridge %s not found.", name)
                sys.exit(1)
            self.phys_to_dpid[phys_net] = name_to_dpid[name]
        LOG.debug("phys_to_dpid: %s", self.phys_to_dpid)

        interfaces = self.lagopus_client.show_interfaces()
        ports = self.lagopus_client.show_ports()
        LOG.debug("interfaces: %s", interfaces)
        LOG.debug("ports: %s", ports)

        # init vhost
        vhost_interfaces = [inter for inter in interfaces
                            if inter["device"].startswith("eth_vhost")]
        self.num_vhost = len(vhost_interfaces)
        used_interfaces = [p["interface"] for p in ports]
        self.used_vhost_id = []
        for inter in vhost_interfaces:
            if inter["name"] in used_interfaces:
                vhost_dev = inter['device'].split(',')[0]
                vhost_id = int(vhost_dev[len("eth_vhost"):])
                self.used_vhost_id.append(vhost_id)
        LOG.debug("num_vhost: %d, used_vhost_id: %s", self.num_vhost,
                  self.used_vhost_id)

        # init pipe
        pipe_interfaces = [inter for inter in interfaces
                           if inter["device"].startswith("eth_pipe")]
        self.num_pipe = len(pipe_interfaces)
        # TODO(hichihara) pipe interface does not remove now.

    def _get_init_bridges(self):
        for retry in range(MAX_WAIT_LAGOPUS_RETRY):
            raw_bridges = self.lagopus_client.show_bridges()
            if raw_bridges:
                LOG.debug("bridges: %s", raw_bridges)
                return raw_bridges
            LOG.debug("Lagopus may not be initialized. waiting")
            eventlet.sleep(10)
        LOG.error("Lagopus isn't running")
        sys.exit(1)

    def get_vhost_interface(self):
        if self.num_vhost == len(self.used_vhost_id):
            # create new vhost interface
            vhost_id = self.num_vhost
            sock_path = "/tmp/sock%d" % vhost_id
            device = "eth_vhost%d,iface=%s" % (vhost_id, sock_path)
            name = "vhost_%d" % vhost_id
            self.lagopus_client.create_vhost_interface(name, device)
            self.num_vhost += 1
            LOG.debug("vhost %d added.", vhost_id)
            os.system("sudo chmod 777 %s" % sock_path)
        else:
            for vhost_id in range(self.num_vhost):
                if vhost_id not in self.used_vhost_id:
                    sock_path = "/tmp/sock%d" % vhost_id
                    name = "vhost_%d" % vhost_id
                    break
        self.used_vhost_id.append(vhost_id)
        return name, sock_path

    def free_vhost_id(self, vhost_id):
        if vhost_id in self.used_vhost_id:
            self.used_vhost_id.remove(vhost_id)

    def port_to_vhost_id(self, port_id):
        ports = self.lagopus_client.show_ports()
        for port in ports:
            if port["name"] == port_id:
                interface = port["interface"]
                if interface.startswith("vhost_"):
                    vhost_id = int(interface[len("vhost_"):])
                    return vhost_id
                return

    def get_pipe(self):
        name0 = "pipe-%d" % self.num_pipe
        name1 = "pipe-%d" % (self.num_pipe + 1)
        device0 = "eth_pipe%d" % self.num_pipe
        device1 = "eth_pipe%d,attach=%s" % (self.num_pipe + 1, device0)
        self.num_pipe += 2

        self.lagopus_client.create_pipe_interface(name0, device0)
        self.lagopus_client.create_pipe_interface(name1, device1)

        return name0, name1

    def get_all_devices(self):
        devices = set()
        ports = self.lagopus_client.show_ports()
        for port in ports:
            devices.add(port["name"])
        LOG.debug("get_all_devices: %s", devices)
        return devices

    def _create_channel(self, channel):
        data = self.lagopus_client.show_channels()
        names = [d['name'] for d in data]
        if channel not in names:
            self.lagopus_client.create_channel(channel)

    def _create_controller(self, controller, channel):
        data = self.lagopus_client.show_controllers()
        names = [d['name'] for d in data]
        if controller not in names:
            self.lagopus_client.create_controller(controller, channel)

    def _create_bridge(self, brname, controller, dpid):
        data = self.lagopus_client.show_bridges()
        names = [d['name'] for d in data]
        if brname not in names:
            self.lagopus_client.create_bridge(brname, controller, dpid)

    def get_bridge(self, segment):
        vlan_id = (segment['segmentation_id']
                   if segment['network_type'] == p_constants.TYPE_VLAN
                   else 0)
        phys_net = segment['physical_network']
        if phys_net not in self.phys_to_dpid:
            # Error
            return
        dpid = (vlan_id << 48) | self.phys_to_dpid[phys_net]
        LOG.debug("vlan_id %d phys dpid %d", vlan_id,
                  self.phys_to_dpid[phys_net])
        LOG.debug("dpid %d 0x%x", dpid, dpid)
        if dpid in self.bridges:
            return self.bridges[dpid]

        # bridge for vlan physical_network does not exist.
        # so create the bridge.
        brname = "%s_%d" % (phys_net, vlan_id)
        channel = "ch-%s" % brname
        self._create_channel(channel)
        controller = "con-%s" % brname
        self._create_controller(controller, channel)
        self._create_bridge(brname, controller, dpid)

        bridge = LagopusBridge(self.ryu_app, brname, dpid, None)
        self.bridges[dpid] = bridge

        pipe1, pipe2 = self.get_pipe()
        port1 = "p-%s" % pipe1
        port2 = "p-%s" % pipe2
        self.lagopus_client.create_port(port1, pipe1)
        self.lagopus_client.create_port(port2, pipe2)

        phys_bridge = self.bridges[self.phys_to_dpid[phys_net]]
        bridge.add_port(port1)
        phys_bridge.add_port(port2)

        phys_bridge.install_vlan(vlan_id, port2)

        return bridge

    @log_helpers.log_method_call
    def plug_vhost(self, context, **kwargs):
        port_id = kwargs['port_id']
        segment = kwargs['segment']

        bridge = self.get_bridge(segment)
        if not bridge:
            # raise
            return

        interface_name, sock_path = self.get_vhost_interface()
        self.lagopus_client.create_port(port_id, interface_name)
        bridge.add_port(port_id)
        return sock_path

    @log_helpers.log_method_call
    def unplug_vhost(self, context, **kwargs):
        port_id = kwargs['port_id']
        bridge_name, _ = self.lagopus_client.find_bridge_port(port_id)
        if not bridge_name:
            LOG.debug("port %s is already unpluged.", port_id)
            return
        self.lagopus_client.bridge_del_port(bridge_name, port_id)
        vhost_id = self.port_to_vhost_id(port_id)
        self.lagopus_client.destroy_port(port_id)
        if vhost_id:
            self.free_vhost_id(vhost_id)

    @log_helpers.log_method_call
    def plug_rawsock(self, context, **kwargs):
        device = kwargs['device']
        segment = kwargs['segment']

        if segment is None:
            LOG.debug("no segment. port may not exist.")
            return

        bridge = self.get_bridge(segment)
        if not bridge:
            return

        interface_name = 'i' + device
        port_name = 'p' + device
        self.lagopus_client.create_rawsock_interface(interface_name, device)
        self.lagopus_client.create_port(port_name, interface_name)
        bridge.add_port(port_name)

        return True

    @log_helpers.log_method_call
    def unplug_rawsock(self, context, **kwargs):
        device = kwargs['device']
        interface_name = 'i' + device
        port_name = 'p' + device

        bridge_name, _ = self.lagopus_client.find_bridge_port(port_name)
        if not bridge_name:
            LOG.debug("device %s is already unpluged.", device)
            return

        self.lagopus_client.bridge_del_port(bridge_name, port_name)
        self.lagopus_client.destroy_port(port_name)
        self.lagopus_client.destroy_interface(interface_name)


class LagopusAgent(service.Service):

    def __init__(self, ryu_app, bridge_mappings, report_interval,
                 quitting_rpc_timeout, agent_type, agent_binary):
        super(LagopusAgent, self).__init__()
        self.ryu_app = ryu_app
        self.bridge_mappings = bridge_mappings
        self.report_interval = report_interval
        self.quitting_rpc_timeout = quitting_rpc_timeout
        self.agent_type = agent_type
        self.agent_binary = agent_binary
        self.state_rpc = agent_rpc.PluginReportStateAPI(topics.REPORTS)

    def start(self):
        self.context = context.get_admin_context_without_session()
        self.manager = LagopusManager(self.ryu_app, self.bridge_mappings)
        self.connection = n_rpc.create_connection()
        self.connection.create_consumer("q-lagopus", [self.manager])

        configurations = {'bridge_mappings': self.bridge_mappings}

        self.agent_state = {
            'binary': self.agent_binary,
            'host': cfg.CONF.host,
            'topic': "q-lagopus",
            'configurations': configurations,
            'agent_type': self.agent_type,
            'resource_versions': resources.LOCAL_RESOURCE_VERSIONS,
            'start_flag': True}

        if self.report_interval:
            heartbeat = loopingcall.FixedIntervalLoopingCall(
                self._report_state)
            heartbeat.start(interval=self.report_interval)

        LOG.info("Agent initialized successfully, now running... ")

        self.connection.consume_in_threads()

    def _report_state(self):
        try:
            devices = len(self.manager.get_all_devices())
            self.agent_state['configurations']['devices'] = devices
            self.state_rpc.report_state(self.context, self.agent_state, True)
            # we only want to update resource versions on startup
            self.agent_state.pop('resource_versions', None)
            self.agent_state.pop('start_flag', None)
        except Exception:
            LOG.exception("Failed reporting state!")

    def stop(self, graceful=True):
        LOG.info("Stopping %s agent.", self.agent_type)
        if graceful and self.quitting_rpc_timeout:
            self.state_rpc.client.timeout = self.quitting_rpc_timeout
        super(LagopusAgent, self).stop(graceful)

    def reset(self):
        common_config.setup_logging()


def parse_bridge_mappings():
    try:
        bridge_mappings = helpers.parse_mappings(
            cfg.CONF.lagopus.bridge_mappings)
        LOG.info("Bridge mappings: %s", bridge_mappings)
        return bridge_mappings
    except ValueError as e:
        LOG.error("Parsing bridge_mappings failed: %s. "
                  "Agent terminated!", e)
        sys.exit(1)


def main(ryu_app):
    bridge_mappings = parse_bridge_mappings()
    report_interval = cfg.CONF.AGENT.report_interval
    quitting_rpc_timeout = cfg.CONF.AGENT.quitting_rpc_timeout
    agent = LagopusAgent(ryu_app, bridge_mappings, report_interval,
                         quitting_rpc_timeout,
                         AGENT_TYPE_LAGOPUS,
                         LAGOPUS_AGENT_BINARY)
    launcher = service.launch(cfg.CONF, agent)
    launcher.wait()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import portbindings
from neutron_lib import constants
from neutron_lib import context as n_context
from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from neutron.plugins.common import constants as p_constants
from neutron.plugins.ml2 import driver_api as api
from neutron.plugins.ml2.drivers import mech_agent

from networking_lagopus.agent import rpc


LOG = logging.getLogger(__name__)
AGENT_TYPE_LAGOPUS = 'Lagopus agent'


class LagopusMechanismDriver(mech_agent.SimpleAgentMechanismDriverBase):

    @log_helpers.log_method_call
    def __init__(self):
        super(LagopusMechanismDriver, self).__init__(
            AGENT_TYPE_LAGOPUS, portbindings.VIF_TYPE_BRIDGE,
            {portbindings.CAP_PORT_FILTER: False})
        self.context = n_context.get_admin_context_without_session()
        self.lagopus_api = rpc.LagopusAgentApi()

    def get_allowed_network_types(self, agent=None):
        return [p_constants.TYPE_FLAT, p_constants.TYPE_VLAN]

    def get_mappings(self, agent):
        return agent['configurations'].get('bridge_mappings', {})

    def try_to_bind_segment_for_agent(self, context, segment, agent):
        if self.check_segment_for_agent(segment, agent):
            vif_type = self.vif_type
            vif_details = dict(self.vif_details)

            if (context.current['device_owner'].
                    startswith(constants.DEVICE_OWNER_COMPUTE_PREFIX)):
                # use vhostuser for VM
                vif_type = portbindings.VIF_TYPE_VHOST_USER
                vif_details[portbindings.VHOST_USER_MODE] = (
                    portbindings.VHOST_USER_MODE_CLIENT)

                sock_path = self.lagopus_api.plug_vhost(
                    self.context, context.current['id'],
                    segment, context._binding.host)
                vif_details[portbindings.VHOST_USER_SOCKET] = sock_path

            context.set_binding(segment[api.ID], vif_type, vif_details,
                                status=constants.PORT_STATUS_ACTIVE)
            return True
        else:
            return False

    @log_helpers.log_method_call
    def update_port_postcommit(self, context):
        if (context.original_host
                and context.original_vif_type == 'vhostuser'
                and not context.host and context.vif_type == 'unbound'):
            self.lagopus_api.unplug_vhost(self.context,
                                          context.current['id'],
                                          context.original_host)

    @log_helpers.log_method_call
    def delete_port_postcommit(self, context):
        if context.host and context.vif_type == 'vhostuser':
            self.lagopus_api.unplug_vhost(self.context,
                                          context.current['id'],
                                          context.host)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_utils import excutils
import ryu.app.ofctl.api  # noqa
from ryu.base import app_manager
from ryu.lib import hub
from ryu.ofproto import ofproto_v1_3

from networking_lagopus.ml2.agent import lagopus_agent


LOG = logging.getLogger(__name__)


def agent_main_wrapper(ryu_app):
    try:
        lagopus_agent.main(ryu_app)
    except Exception:
        with excutils.save_and_reraise_exception():
            LOG.exception("Agent main thread died of an exception")
    finally:
        # The following call terminates Ryu's AppManager.run_apps(),
        # which is needed for clean shutdown of an agent process.
        # The close() call must be called in another thread, otherwise
        # it suicides and ends prematurely.
        hub.spawn(app_manager.AppManager.get_instance().close)


class LagopusAgentRyuApp(app_manager.RyuApp):
    OFP_VERSIONS = [ofproto_v1_3.OFP_VERSION]

    def start(self):
        # Start Ryu event loop thread
        super(LagopusAgentRyuApp, self).start()
        self.threads.append(hub.spawn(agent_main_wrapper, self,
                                      raise_error=True))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys


LAGOPUS_DSL_TEMPLATE = "lagopus_template.dsl"


def main():
    if len(sys.argv) < 3:
        print("usage: generate_dsl dpdk_port_mappings dsl_path")
        return 1
    dpdk_port_mappings = sys.argv[1]
    dsl_path = sys.argv[2]

    phys_nets = []
    for map in dpdk_port_mappings.split(','):
        _, phys = map.split('#')
        phys_nets.append(phys)

    with open(LAGOPUS_DSL_TEMPLATE, "r") as dsl:
        bridge_conf = dsl.read()

    with open(dsl_path, "w") as f:
        num = 1
        for phys in phys_nets:
            f.write(bridge_conf % {'num': num, 'port': num - 1})
            num += 1


if __name__ == "__main__":
    sys.exit(main())



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging

from neutron.common import rpc as n_rpc


LOG = logging.getLogger(__name__)


class LagopusAgentApi(object):
    '''Lagopus agent RPC API

    API version history:
        1.0 - Initial version.
    '''

    def __init__(self):
        target = oslo_messaging.Target(version='1.0')
        self.client = n_rpc.get_client(target)

    def _get_context(self, host):
        if host is None:
            host = cfg.CONF.host
        topic = "q-lagopus.%s" % host
        return self.client.prepare(topic=topic)

    def plug_rawsock(self, context, device, segment, host=None):
        cctxt = self._get_context(host)
        return cctxt.call(context, 'plug_rawsock',
                          device=device, segment=segment)

    def unplug_rawsock(self, context, device, host=None):
        cctxt = self._get_context(host)
        return cctxt.call(context, 'unplug_rawsock',
                          device=device)

    def plug_vhost(self, context, port_id, segment, host=None):
        cctxt = self._get_context(host)
        return cctxt.call(context, 'plug_vhost',
                          port_id=port_id, segment=segment)

    def unplug_vhost(self, context, port_id, host=None):
        cctxt = self._get_context(host)
        return cctxt.call(context, 'unplug_vhost',
                          port_id=port_id)
